@@ -22,7 +22,7 @@ import extras
 
 
 def filter_by_ids(suite_or_case, test_ids):
-    """Return a test suite with the intersection of suite_or_case and test_ids.
+    """Remove tests from suite_or_case where their id is not in test_ids.
 
     :param suite_or_case: A test suite or test case.
     :param test_ids: Something that supports the __contains__ protocol.
@@ -58,31 +58,23 @@ def filter_by_ids(suite_or_case, test_ids):
     thus the backwards-compatible code paths attempt to mutate in place rather
     than guessing how to reconstruct a new suite.
     """
-    def _filter_tests(to_filter, id_map):
-        # Compatible objects
-        if extras.safe_hasattr(to_filter, 'filter_by_ids'):
-            res = to_filter.filter_by_ids(test_ids)
-            _filter_tests(res, id_map)
-        # TestCase objects.
-        elif extras.safe_hasattr(to_filter, 'id'):
-            test_id = to_filter.id()
-            if test_id in test_ids:
-                id_map[test_id] = to_filter
-        # Standard TestSuites or derived classes.
-        elif isinstance(to_filter, unittest.TestSuite):
-            for item in to_filter:
-                _filter_tests(item, id_map)
-
-    id_map = {}
-    _filter_tests(suite_or_case, id_map)
-
-    result = unittest.TestSuite()
-    for test_id in test_ids:
-        if test_id in id_map:
-            # Use pop to avoid duplicates
-            result.addTest(id_map.pop(test_id))
-
-    return result
+    # Compatible objects
+    if extras.safe_hasattr(suite_or_case, "filter_by_ids"):
+        return suite_or_case.filter_by_ids(test_ids)
+    # TestCase objects.
+    if extras.safe_hasattr(suite_or_case, "id"):
+        if suite_or_case.id() in test_ids:
+            return suite_or_case
+        else:
+            return unittest.TestSuite()
+    # Standard TestSuites or derived classes [assumed to be mutable].
+    if isinstance(suite_or_case, unittest.TestSuite):
+        filtered = []
+        for item in suite_or_case:
+            filtered.append(filter_by_ids(item, test_ids))
+        suite_or_case._tests[:] = filtered
+    # Everything else:
+    return suite_or_case
 
 
 def iterate_tests(test_suite_or_case):
@@ -111,17 +103,18 @@ def list_test(test):
         describing things that failed to import.
     """
     unittest_import_strs = {
-        'unittest2.loader.ModuleImportFailure.',
-        'unittest.loader.ModuleImportFailure.',
-        'discover.ModuleImportFailure.'
-        }
+        "unittest2.loader.ModuleImportFailure.",
+        "unittest.loader.ModuleImportFailure.",
+        "discover.ModuleImportFailure.",
+        "unittest.loader._FailedTest.",
+    }
     test_ids = []
     errors = []
     for test in iterate_tests(test):
         # Much ugly.
         for prefix in unittest_import_strs:
             if test.id().startswith(prefix):
-                errors.append(test.id()[len(prefix):])
+                errors.append(test.id()[len(prefix) :])
                 break
         else:
             test_ids.append(test.id())
@@ -136,13 +129,24 @@ class TestProgram(unittest.TestProgram):
     failfast = catchbreak = buffer = progName = None
     _discovery_parser = None
 
-    def __init__(self, module='__main__', defaultTest=None, argv=None,
-                 testRunner=None, testLoader=unittest.defaultTestLoader,
-                 exit=False, verbosity=1, failfast=None, catchbreak=None,
-                 buffer=None, warnings=None, tb_locals=False):
+    def __init__(
+        self,
+        module="__main__",
+        defaultTest=None,
+        argv=None,
+        testRunner=None,
+        testLoader=unittest.defaultTestLoader,
+        exit=False,
+        verbosity=1,
+        failfast=None,
+        catchbreak=None,
+        buffer=None,
+        warnings=None,
+        tb_locals=False,
+    ):
         if isinstance(module, str):
             self.module = __import__(module)
-            for part in module.split('.')[1:]:
+            for part in module.split(".")[1:]:
                 self.module = getattr(self.module, part)
         else:
             self.module = module
@@ -159,7 +163,7 @@ class TestProgram(unittest.TestProgram):
             # even if DeprecationWarnings are ignored by default
             # print them anyway unless other warnings settings are
             # specified by the warnings arg or the -W python flag
-            self.warnings = 'default'
+            self.warnings = "default"
         else:
             # here self.warnings is set either to the value passed
             # to the warnings args or to None.
@@ -181,45 +185,53 @@ class TestProgram(unittest.TestProgram):
             # does in OptimisingTestSuite.add, but with a standard protocol).
             # This is needed because the load_tests hook allows arbitrary
             # suites, even if that is rarely used.
-            source = open(self.load_list, 'rb')
+            source = open(self.load_list, "rb")
             try:
                 lines = source.readlines()
             finally:
                 source.close()
-            test_ids = [line.strip().decode('utf-8') for line in lines]
+            test_ids = {line.strip().decode("utf-8") for line in lines}
             self.test = filter_by_ids(self.test, test_ids)
-
         # XXX: Local edit (see http://bugs.python.org/issue22860)
         if not self.listtests:
             self.runTests()
         else:
             runner = self._get_runner()
-            if extras.safe_hasattr(runner, 'list'):
+            if extras.safe_hasattr(runner, "list"):
                 try:
                     runner.list(self.test, loader=self.testLoader)
                 except TypeError:
                     runner.list(self.test)
             else:
                 for test in iterate_tests(self.test):
-                    sys.stdout.write('%s\n' % test.id())
+                    sys.stdout.write("%s\n" % test.id())
 
     def _getParentArgParser(self):
         parser = super(TestProgram, self)._getParentArgParser()
         # XXX: Local edit (see http://bugs.python.org/issue22860)
         parser.add_argument(
-            '-l', '--list', dest='listtests', default=False,
-            action='store_true', help='List tests rather than executing them')
+            "-l",
+            "--list",
+            dest="listtests",
+            default=False,
+            action="store_true",
+            help="List tests rather than executing them",
+        )
         parser.add_argument(
-            '--load-list', dest='load_list', default=None,
-            help='Specifies a file containing test ids, only tests matching '
-                 'those ids are executed')
+            "--load-list",
+            dest="load_list",
+            default=None,
+            help="Specifies a file containing test ids, only tests matching "
+            "those ids are executed",
+        )
         return parser
 
     def _get_runner(self):
         testRunner = self.testRunner
         try:
-            testRunner = self.testRunner(failfast=self.failfast,
-                                         tb_locals=self.tb_locals)
+            testRunner = self.testRunner(
+                failfast=self.failfast, tb_locals=self.tb_locals
+            )
         except TypeError:
             testRunner = self.testRunner()
         # If for some reason we failed to initialize the runner initialize
